/*
 * Supervisor Mode Access Prevention support
 *
 * Copyright (C) 2012 Intel Corporation
 * Author: H. Peter Anvin <hpa@linux.intel.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 */

#ifndef _ASM_X86_SMAP_H
#define _ASM_X86_SMAP_H

#include <asm/nops.h>
#include <asm/cpufeatures.h>
#include <asm/nospec-branch.h>

#if defined(CONFIG_X86_64) && defined(CONFIG_PAX_MEMORY_UDEREF)
#define ASM_PAX_OPEN_USERLAND "pax_direct_call __pax_open_userland"
#define ASM_PAX_CLOSE_USERLAND "pax_direct_call __pax_close_userland"
#else
#define ASM_PAX_OPEN_USERLAND ""
#define ASM_PAX_CLOSE_USERLAND ""
#endif

/* "Raw" instruction opcodes */
#define __ASM_CLAC	".byte 0x0f,0x01,0xca"
#define __ASM_STAC	".byte 0x0f,0x01,0xcb"

#ifdef __ASSEMBLY__

#ifdef CONFIG_X86_SMAP

#define ASM_CLAC \
	ALTERNATIVE "", __ASM_CLAC, X86_FEATURE_SMAP

#define ASM_PAX_CLAC __ASM_CLAC

#define ASM_STAC \
	ALTERNATIVE "", __ASM_STAC, X86_FEATURE_SMAP

#define ASM_PAX_STAC __ASM_STAC

#else /* CONFIG_X86_SMAP */

#define ASM_CLAC
#define ASM_STAC

#define ASM_PAX_CLAC ""
#define ASM_PAX_STAC ""

#endif /* CONFIG_X86_SMAP */

#define ASM_USER_ACCESS_BEGIN	ALTERNATIVE_2 \
	"", \
	ASM_PAX_OPEN_USERLAND, \
	X86_FEATURE_UDEREF, \
	ASM_PAX_STAC, \
	X86_FEATURE_SMAP

#define ASM_USER_ACCESS_END	ALTERNATIVE_2 \
	"", \
	ASM_PAX_CLOSE_USERLAND, \
	X86_FEATURE_UDEREF, \
	ASM_PAX_CLAC, \
	X86_FEATURE_SMAP

#else /* __ASSEMBLY__ */

#include <asm/alternative.h>

#define __HAVE_ARCH_PAX_OPEN_USERLAND
#define __HAVE_ARCH_PAX_CLOSE_USERLAND

#if defined(CONFIG_X86_64) && defined(CONFIG_PAX_MEMORY_UDEREF)
#ifdef CONFIG_CC_IS_CLANG
#define __preserve_all_registers __no_caller_saved_registers
#else
#define __preserve_all_registers /* handled via -fcall-saved-* in CFLAGS */
#endif

void __used __preserve_all_registers __pax_open_userland(void);
void __used __preserve_all_registers __pax_close_userland(void);
#define PAX_OPEN_USERLAND PAX_DIRECT_CALL("%P[open]")
#define PAX_OPEN_USERLAND_CONSTRAINTS : ASM_CALL_CONSTRAINT : [open] "i" (__pax_open_userland) : "memory"
#define PAX_CLOSE_USERLAND PAX_DIRECT_CALL("%P[close]")
#define PAX_CLOSE_USERLAND_CONSTRAINTS : ASM_CALL_CONSTRAINT : [close] "i" (__pax_close_userland) : "memory"
#else
#define PAX_OPEN_USERLAND ""
#define PAX_OPEN_USERLAND_CONSTRAINTS : : : "memory"
#define PAX_CLOSE_USERLAND ""
#define PAX_CLOSE_USERLAND_CONSTRAINTS : : : "memory"
#endif

#ifdef CONFIG_X86_SMAP

#define CLAC __ASM_CLAC
#define STAC __ASM_STAC

static __always_inline unsigned long smap_save(void)
{
	unsigned long flags;

	asm volatile (ALTERNATIVE("", "pushf; pop %0; " __ASM_CLAC,
				  X86_FEATURE_SMAP)
		      : "=rm" (flags) : : "memory", "cc");

	return flags;
}

static __always_inline void smap_restore(unsigned long flags)
{
	asm volatile (ALTERNATIVE("", "push %0; popf", X86_FEATURE_SMAP)
		      : : "g" (flags) : "memory", "cc");
}

/* These macros can be used in asm() statements */
#define ASM_CLAC \
	ALTERNATIVE("", __ASM_CLAC, X86_FEATURE_SMAP)
#define ASM_STAC \
	ALTERNATIVE("", __ASM_STAC, X86_FEATURE_SMAP)

#else /* CONFIG_X86_SMAP */

#define CLAC ""
#define STAC ""

static inline unsigned long smap_save(void) { return 0; }
static inline void smap_restore(unsigned long flags) { }

#define ASM_CLAC
#define ASM_STAC

#endif /* CONFIG_X86_SMAP */

#endif /* __ASSEMBLY__ */

#endif /* _ASM_X86_SMAP_H */
