#ifndef __LINUX_COMPILER_TYPES_H
#define __LINUX_COMPILER_TYPES_H

#ifndef __ASSEMBLY__

#ifdef __CHECKER__
# define __user		__attribute__((noderef, address_space(1)))
# define __force_user	__force __user
# define __kernel	__attribute__((address_space(0)))
# define __force_kernel	__force __kernel
# define __safe		__attribute__((safe))
# define __force	__attribute__((force))
# define __nocast	__attribute__((nocast))
# define __iomem	__attribute__((noderef, address_space(2)))
# define __force_iomem	__force __iomem
# define __must_hold(x)	__attribute__((context(x,1,1)))
# define __acquires(x)	__attribute__((context(x,0,1)))
# define __releases(x)	__attribute__((context(x,1,0)))
# define __acquire(x)	__context__(x,1)
# define __release(x)	__context__(x,-1)
# define __cond_lock(x,c)	((c) ? ({ __acquire(x); 1; }) : 0)
# define __opens(x)	__attribute__((context(#x,0,1)))
# define __closes(x)	__attribute__((context(#x,1,0)))
# define __open(x)	__context__(#x,1)
# define __close(x)	__context__(#x,-1)
# define __percpu	__attribute__((noderef, address_space(3)))
# define __force_percpu	__force __percpu
# define __rcu		__attribute__((noderef, address_space(4)))
# define __force_rcu	__force __rcu
# define __private	__attribute__((noderef))
extern void __chk_user_ptr(const volatile void __user *);
extern void __chk_io_ptr(const volatile void __iomem *);
# define ACCESS_PRIVATE(p, member) (*((typeof((p)->member) __force *) &(p)->member))
#else /* __CHECKER__ */
# ifdef CHECKER_PLUGIN
#  ifdef CHECKER_PLUGIN_USER
//#  define __user
//#  define __force_user
//#  define __kernel
//#  define __force_kernel
#  else /* CHECKER_PLUGIN_USER */
#   define __user
#   define __force_user
#   define __kernel
#   define __force_kernel
#  endif /* CHECKER_PLUGIN_USER */
#  ifdef CHECKER_PLUGIN_CONTEXT
#   define __must_hold(x)	__attribute__((context(#x,1,1)))
#   define __acquires(x)	__attribute__((context(#x,0,1)))
#   define __releases(x)	__attribute__((context(#x,1,0)))
#   define __acquire(x)		__context__(#x,1)
#   define __release(x)		__context__(#x,-1)
#   define __cond_lock(x,c)	((c) ? ({ __acquire(x); 1; }) : 0)
#   define __cond_unlock(x,c)	((c) ? ({ __release(x); 1; }) : 0)
#  else /* CHECKER_PLUGIN_CONTEXT */
#   define __must_hold(x)
#   define __acquires(x)
#   define __releases(x)
#   define __acquire(x) (void)0
#   define __release(x) (void)0
#   define __cond_lock(x,c) (c)
#   define __cond_unlock(x,c) (c)
#  endif /* CHECKER_PLUGIN_CONTEXT */
# else /* CHECKER_PLUGIN */
#  ifdef STRUCTLEAK_PLUGIN
#   define __user __attribute__((user))
#  else
#   define __user
#  endif
#  define __force_user
#  define __kernel
#  define __force_kernel
#  if 1
#   define __must_hold(x)
#   define __acquires(x)
#   define __releases(x)
#   define __acquire(x) (void)0
#   define __release(x) (void)0
#   define __cond_lock(x,c) (c)
#  else
#   define __must_hold(x)	__attribute__((context(#x,1,1)))
#   define __acquires(x)	__attribute__((context(#x,0,1)))
#   define __releases(x)	__attribute__((context(#x,1,0)))
#   define __acquire(x)		__context__(#x,1)
#   define __release(x)		__context__(#x,-1)
#   define __cond_lock(x,c)	((c) ? ({ __acquire(x); 1; }) : 0)
#   define __cond_unlock(x,c)	((c) ? ({ __release(x); 1; }) : 0)
# endif
# endif /* CHECKER_PLUGIN */
# define __safe
# define __force
# define __nocast
# define __iomem
# define __force_iomem
# define __chk_user_ptr(x) (void)0
# define __chk_io_ptr(x) (void)0
# define __builtin_warning(x, y...) (1)
# define __percpu
# define __force_percpu
# define __rcu
# define __force_rcu
# define __private
# define ACCESS_PRIVATE(p, member) ((p)->member)
# ifdef CONSTIFY_PLUGIN
#  define __opens(x)	__attribute__((context(#x,0,1)))
#  define __closes(x)	__attribute__((context(#x,1,0)))
#  define __open(x)	__context__(#x,1)
#  define __close(x)	__context__(#x,-1)
# else /* CONSTIFY_PLUGIN */
#  define __opens(x)
#  define __closes(x)
#  define __open(x)	(void)0
#  define __close(x)	(void)0
# endif /* CONSTIFY_PLUGIN */
#endif /* __CHECKER__ */

#ifdef __KERNEL__


/*
 * Some architectures need to provide custom definitions of macros provided
 * by linux/compiler-*.h, and can do so using asm/compiler.h. We include that
 * conditionally rather than using an asm-generic wrapper in order to avoid
 * build failures if any C compilation, which will include this file via an
 * -include argument in c_flags, occurs prior to the asm-generic wrappers being
 * generated.
 */
#ifdef CONFIG_HAVE_ARCH_COMPILER_H
#include <asm/compiler.h>
#endif

#ifndef RAP_SAFE_ASM
#define RAP_SAFE_ASM
#endif

#ifdef __GNUC__
#include <linux/compiler-gcc.h>
#endif

#if defined(CC_USING_HOTPATCH) && !defined(__CHECKER__)
#define notrace __attribute__((hotpatch(0,0)))
#else
#define notrace __attribute__((no_instrument_function))
#endif

/* Intel compiler defines __GNUC__. So we will overwrite implementations
 * coming from above header files here
 */
#ifdef __INTEL_COMPILER
# include <linux/compiler-intel.h>
#endif

/* Clang compiler defines __GNUC__. So we will overwrite implementations
 * coming from above header files here
 */
#ifdef __clang__
#include <linux/compiler-clang.h>
#endif

/*
 * Generic compiler-dependent macros required for kernel
 * build go below this comment. Actual compiler/compiler version
 * specific implementations come from the above header files
 */

struct ftrace_branch_data {
	const char *func;
	const char *file;
	unsigned line;
	union {
		struct {
			unsigned long correct;
			unsigned long incorrect;
		};
		struct {
			unsigned long miss;
			unsigned long hit;
		};
		unsigned long miss_hit[2];
	};
};

struct ftrace_likely_data {
	struct ftrace_branch_data	data;
	unsigned long			constant;
};

#endif /* __KERNEL__ */

#endif /* __ASSEMBLY__ */

#ifdef __KERNEL__
/*
 * Allow us to mark functions as 'deprecated' and have gcc emit a nice
 * warning for each use, in hopes of speeding the functions removal.
 * Usage is:
 * 		int __deprecated foo(void)
 */
#ifndef __deprecated
# define __deprecated		/* unimplemented */
#endif

#ifdef MODULE
#define __deprecated_for_modules __deprecated
#else
#define __deprecated_for_modules
#endif

#ifndef __must_check
#define __must_check
#endif

#ifndef CONFIG_ENABLE_MUST_CHECK
#undef __must_check
#define __must_check
#endif
#ifndef CONFIG_ENABLE_WARN_DEPRECATED
#undef __deprecated
#undef __deprecated_for_modules
#define __deprecated
#define __deprecated_for_modules
#endif

#ifndef __malloc
#define __malloc
#endif

/*
 * Allow us to avoid 'defined but not used' warnings on functions and data,
 * as well as force them to be emitted to the assembly file.
 *
 * As of gcc 3.4, static functions that are not marked with attribute((used))
 * may be elided from the assembly file.  As of gcc 3.4, static data not so
 * marked will not be elided, but this may change in a future gcc version.
 *
 * NOTE: Because distributions shipped with a backported unit-at-a-time
 * compiler in gcc 3.3, we must define __used to be __attribute__((used))
 * for gcc >=3.3 instead of 3.4.
 *
 * In prior versions of gcc, such functions and data would be emitted, but
 * would be warned about except with attribute((unused)).
 *
 * Mark functions that are referenced only in inline assembly as __used so
 * the code is emitted even though it appears to be unreferenced.
 */
#ifndef __used
# define __used			/* unimplemented */
#endif

#ifndef __maybe_unused
# define __maybe_unused		/* unimplemented */
#endif

#ifndef __always_unused
# define __always_unused	/* unimplemented */
#endif

#ifndef noinline
#define noinline
#endif

/*
 * Rather then using noinline to prevent stack consumption, use
 * noinline_for_stack instead.  For documentation reasons.
 */
#define noinline_for_stack noinline

/* Disable inlining especially under LTO that would otherwise lead to autoslab name collisions */
#define noinline_for_autoslab noinline

#ifndef __always_inline
#define __always_inline inline
#endif

/* Indirect macros required for expanded argument pasting, eg. __LINE__. */
#define ___PASTE(a,b) a##b
#define __PASTE(a,b) ___PASTE(a,b)

#endif /* __KERNEL__ */

/*
 * From the GCC manual:
 *
 * Many functions do not examine any values except their arguments,
 * and have no effects except the return value.  Basically this is
 * just slightly more strict class than the `pure' attribute above,
 * since function is not allowed to read global memory.
 *
 * Note that a function that has pointer arguments and examines the
 * data pointed to must _not_ be declared `const'.  Likewise, a
 * function that calls a non-`const' function usually must not be
 * `const'.  It does not make sense for a `const' function to return
 * `void'.
 */
#ifndef __attribute_const__
# define __attribute_const__	/* unimplemented */
#endif

#ifndef __designated_init
# define __designated_init
#endif

#ifndef __latent_entropy
# define __latent_entropy
#endif

#ifndef __copy
# define __copy(symbol)
#endif

#ifndef __randomize_layout
# define __randomize_layout __designated_init
#endif

#ifndef __no_randomize_layout
# define __no_randomize_layout
#endif

#ifndef __no_const
# define __no_const
#endif

#ifndef __do_const
# define __do_const
#endif

#ifndef __mutable_const
# define __mutable_const
#endif

#ifndef __size_overflow
# define __size_overflow(...)
#endif

#ifndef __intentional_overflow
# define __intentional_overflow(...)
# define __turn_off_size_overflow
# define __skip_size_overflow
#endif

#ifndef __nocapture
# define __nocapture(...)
#endif

#ifndef __unverified_nocapture
# define __unverified_nocapture(...)
#endif

#ifndef __rap_hash
#define __rap_hash
#endif

/*
 * Tell gcc if a function is cold. The compiler will assume any path
 * directly leading to the call is unlikely.
 */

#ifndef __cold
#define __cold
#endif

#ifndef __rap_hash_override
#define __rap_hash_override(...)
#endif

#ifndef __respectre_inline
#define __respectre_inline __always_inline
#endif

#ifndef __read_once
#define __read_once
#endif

#ifndef __write_once
#define __write_once
#endif

#ifndef __section_type
#define __section_type(...)
#endif

#ifndef __alloc_size
#define __alloc_size(...)
#endif

#ifndef __alloc_gfp_flags
#define __alloc_gfp_flags(...)
#endif

#ifndef __alloc_xfs_km_flags
#define __alloc_xfs_km_flags(...)
#endif

#ifndef __alloc
#define __alloc
#endif

#ifndef __bos
#define __bos(ptr, arg)
#endif

#ifndef __bos0
#define __bos0(ptr)
#endif

#ifndef __bos1
#define __bos1(ptr)
#endif

#ifndef __no_caller_saved_registers
#define __no_caller_saved_registers
#endif

/* Simple shorthand for a section definition */
#ifndef __section
# define __section(S) __attribute__ ((__section__(#S)))
#endif

#ifndef __visible
#define __visible
#endif

#ifndef __nostackprotector
# define __nostackprotector
#endif

/*
 * Assume alignment of return value.
 */
#ifndef __assume_aligned
#define __assume_aligned(a, ...)
#endif


/* Are two types/vars the same type (ignoring qualifiers)? */
#ifndef __same_type
# define __same_type(a, b) __builtin_types_compatible_p(typeof(a), typeof(b))
#endif

#define __type_is_unsigned(t) (__same_type((t)0, 0UL) || __same_type((t)0, 0U) || __same_type((t)0, (unsigned short)0) || __same_type((t)0, (unsigned char)0))

#define ____inttypeof(t, v)	\
	__builtin_choose_expr(__builtin_types_compatible_p(typeof(v), t), (t)0,

#define __inttypeof(t, v)	\
	____inttypeof(signed t, v)\
	____inttypeof(unsigned t, v)

#define inttypeof(v)	typeof(\
	__inttypeof(char, v) \
	__inttypeof(short, v) \
	__inttypeof(int, v) \
	__inttypeof(long, v) \
	__inttypeof(long long, v) \
	(void)0)))))))))))

/* Is this type a native word size -- useful for atomic operations */
#ifndef __native_word
# define __native_word(t) (sizeof(t) == sizeof(char) || sizeof(t) == sizeof(short) || sizeof(t) == sizeof(int) || sizeof(t) == sizeof(long))
#endif

#endif /* __LINUX_COMPILER_TYPES_H */
